/* support for format statements -- called from cmn0.lisp */

#include <stdio.h>

static FILE *fp;

int OpenCMNFile (char *fileName)
{
  fp = fopen(fileName,"w");
  if (fp == NULL) 
    {
      printf("Cannot open %s.\n",fileName);
      fflush(stdout);
      return 1;
    }
return 0;
}

void CloseCMNFile (void)
{
  if (fp != NULL) fclose(fp);
  fp = NULL;
}


void PrintCMNFile (char *arr)
{
    fprintf(fp,"%s\n",arr);
}

void CommentCMNFile (char *arr)
{
    fprintf(fp,"%% %s\n",arr);
}

void LinetoCMNFile (float *arr)
{
    fprintf(fp," %.2f %.2f lineto\n",arr[0],arr[1]);
}

void RLinetoCMNFile (float *arr)
{
    fprintf(fp," %.2f %.2f rlineto\n",arr[0],arr[1]);
}

void MovetoCMNFile (float *arr)
{
    fprintf(fp," %.2f %.2f moveto\n",arr[0],arr[1]);
}

void RMovetoCMNFile (float *arr)
{
    fprintf(fp," %.2f %.2f rmoveto\n",arr[0],arr[1]);
}

void LinewidthCMNFile (float *arr)
{
    fprintf(fp," %.2f setlinewidth\n",arr[0]);
}

void LinewidthClearedCMNFile (float *arr)
{
    fprintf(fp," %.2f setlinewidth stroke 0 setlinewidth\n",arr[0]);
}

void CurvetoCMNFile (float *arr)
{
    fprintf(fp," %.2f %.2f %.2f %.2f %.2f %.2f curveto\n",arr[0],arr[1],arr[2],arr[3],arr[4],arr[5]);
}

void MatrixCMNFile (float *arr)
{
    fprintf(fp," gsave [%.2f %.2f %.2f %.2f %.2f %.2f] concat\n",arr[0],arr[1],arr[2],arr[3],arr[4],arr[5]);
}

void FillinCMNFile (float *arr, char *cmd)
{
  fprintf(fp," %.2f %.2f %.2f %.2f %s\n",arr[0],arr[1],arr[2],arr[3],cmd);
}

void FontTextCMNFile (char *font, int size, char *str)
{
  fprintf(fp," /%s findfont %d scalefont setfont (%s) show\n",font,size,str);
}


#if LINUX && ACL
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#define SND_VERSION      "SND_VERSION"
#define SND_COMMAND      "SND_COMMAND"

static Display *dpy = NULL;
static Window wn;

static Window compare_window(Display *display, Window window, char *id)
{
  Atom type;
  int format;
  unsigned long nitems, bytesafter;
  unsigned char *version[1];
  Window found = (Window)None;
  if (((XGetWindowProperty(display, window, XInternAtom (display, id, False), 0L, (long)BUFSIZ, False,
			   XA_STRING, &type, &format, &nitems, &bytesafter, 
			   (unsigned char **)version)) == Success) && (type != None))
    {
      found = window;
      if (version[0]) 
	XFree((char *)(version[0]));
    }
  return(found);
}

static Window find_window(Display *display, 
			  Window starting_window, 
			  char *name, 
			  Window (*compare_func)(Display *dpy, Window win, char *id))
{
  Window rootwindow, window_parent;
  unsigned int i = 0, num_children = 0;
  Window *children = NULL;
  Window window = (compare_func)(display, starting_window, name);
  if (window != (Window)None) return (window);
  if ((XQueryTree(display, starting_window, &rootwindow, &window_parent, &children, &num_children)) == 0) 
    return ((Window)None);
  while ((i < num_children) && (window == (Window)None))
    window = find_window(display, children[i++], name, compare_func);
  if (children) 
    XFree((char *)children);
  return(window);
}

static Atom cmn_v, cmn_c;
void cmn_init_x(void)
{
  dpy = XOpenDisplay(NULL);
  wn = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 0, 0, 10, 10, 1, CopyFromParent, CopyFromParent);
  cmn_v = XInternAtom(dpy, "CMN_VERSION", False);
  cmn_c = XInternAtom(dpy, "CMN_COMMAND", False);
  XChangeProperty(dpy, wn, cmn_v, XA_STRING, 8, PropModeReplace, "1.0", 4);
}

void cmn_send_snd(char *command)
{
  Window window;
  if (dpy == NULL) 
    cmn_init_x();
  if ((window = find_window(dpy, DefaultRootWindow(dpy), SND_VERSION, compare_window)))
    {
      XChangeProperty(dpy, 
		      window, 
		      XInternAtom(dpy, SND_COMMAND, False), 
		      XA_STRING, 8, 
		      PropModeReplace, 
		      (unsigned char *)command, 
		      strlen(command) + 1);
      XFlush(dpy);
    }
}

static char lisp_buf[4096];
char *cmn_receive_snd(void)
{
  int format;
  Atom type;
  unsigned long nitems, bytesafter;
  unsigned char *version[1];
  if (dpy == NULL) 
    cmn_init_x();
  if (((XGetWindowProperty(dpy, wn, cmn_c, 0L, (long)BUFSIZ, False,
			   XA_STRING, &type, &format, &nitems, &bytesafter, (unsigned char **)version)) == Success) && 
      (type != None))
    {
      if (version[0]) 
	{
	  strcpy(lisp_buf, (char *)(version[0]));
	  XFree(version[0]);
	  XChangeProperty(dpy, wn, cmn_c, XA_STRING, 8, PropModeReplace, "", 1);
	  XFlush(dpy);
	  return(lisp_buf);
	}
    }
  return("");
}

/* (change-property "CMN_VERSION" "CMN_COMMAND" "(print 32)") */
#endif

