/* pipe communications to xcmnw program */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>

#define CLEAR 0
#define LINE 1
#define TEXT 2
#define RECTANGLE 3
#define POLYGON 4
#define CIRCLE 5
#define QUIT 6
#define COLOR 7
#define LINE_WIDTH 8
#define LINE_STYLE 9
#define FILL_STYLE 10

static char *buffer = NULL;
static int buffer_size = 0;
static int *ibuf;
static FILE *fd;

void cmn_popen(char *name)
{
  buffer_size = 8192;
  buffer = (char *)calloc(buffer_size,sizeof(char));
  fd = popen(name,"w");
}

int cmn_pclose(void)
{
  ibuf = (int *)buffer;
  ibuf[0] = 4;
  ibuf[1] = QUIT;
  fwrite(buffer,1,8,fd);
  fflush(fd); 
  return(pclose(fd));
}

void x_clear_window (void)
{
  ibuf = (int *)buffer;
  ibuf[0] = 4;
  ibuf[1] = CLEAR;
  fwrite(buffer,1,8,fd);
  fflush(fd); 
}

void x_color(int r, int g, int b)
{
  ibuf = (int *)buffer;
  ibuf[0] = 16;
  ibuf[1] = COLOR;
  ibuf[2] = r;
  ibuf[3] = g;
  ibuf[4] = b;
  fwrite(buffer,1,20,fd);
  fflush(fd); 
}

void x_line_width(int r)
{
  ibuf = (int *)buffer;
  ibuf[0] = 8;
  ibuf[1] = LINE_WIDTH;
  ibuf[2] = r;
  fwrite(buffer,1,12,fd);
  fflush(fd); 
}

void x_line_style(int r)
{
  ibuf = (int *)buffer;
  ibuf[0] = 8;
  ibuf[1] = LINE_STYLE;
  ibuf[2] = r;
  fwrite(buffer,1,12,fd);
  fflush(fd); 
}

void x_fill_style(int r)
{
  ibuf = (int *)buffer;
  ibuf[0] = 8;
  ibuf[1] = FILL_STYLE;
  ibuf[2] = r;
  fwrite(buffer,1,12,fd);
  fflush(fd); 
}

void x_lineto(int *vals)
{
  int i;
  ibuf = (int *)buffer;
  ibuf[0] = 20;
  ibuf[1] = LINE;
  for (i=0;i<4;i++) ibuf[i+2] = vals[i];
  fwrite(buffer,1,24,fd);
  fflush(fd); 
}

void x_rectangle(int filled, int *vals)
{
  int i;
  ibuf = (int *)buffer;
  ibuf[0] = 24;
  ibuf[1] = RECTANGLE;
  ibuf[2] = filled;
  for (i=0;i<4;i++) ibuf[i+3] = vals[i];
  fwrite(buffer,1,28,fd);
  fflush(fd);
}

void x_circle(int filled, int *vals)
{
  int i;
  ibuf = (int *)buffer;
  ibuf[0] = 28;
  ibuf[1] = CIRCLE;
  ibuf[2] = filled;
  for (i=0;i<5;i++) ibuf[i+3] = vals[i];
  fwrite(buffer,1,32,fd);
  fflush(fd);
}

void x_polygon(int filled, int even_odd, int size, int *vals)
{
  /* size = num pts => 1/2 ints */
  int i,j,bytes;
  bytes = (size*2 + 8) * 4;
  if (bytes > buffer_size)
    {
      buffer_size = bytes*2;
      buffer = (char *)realloc(buffer,buffer_size*sizeof(char));
    }
  ibuf = (int *)buffer;
  ibuf[0] = bytes-4;
  ibuf[1] = POLYGON;
  ibuf[2] = filled;
  ibuf[3] = even_odd;
  ibuf[4] = size;
  for (i=0,j=0;i<size;i++,j+=2)
    {
      ibuf[j+5] = vals[j];
      ibuf[j+6] = vals[j+1];
    }
  fwrite(buffer,1,bytes,fd);
  fflush(fd);
}

void x_text(int size, int *vals, char *font, char *text)
{
  int font_len,text_len,i;
  font_len = strlen(font);
  text_len = strlen(text);
  ibuf = (int *)buffer;
  ibuf[0] = 20 + font_len + 1 + text_len + 1 - 4;
  ibuf[1] = TEXT;
  ibuf[2] = size;
  ibuf[3] = vals[0];
  ibuf[4] = vals[1];
  for (i=0;i<font_len;i++) buffer[20+i] = font[i];
  buffer[20+font_len] = 0;
  for (i=0;i<text_len;i++) buffer[20+font_len+1+i] = text[i];
  buffer[20+font_len+1+text_len] = 0;
  fwrite(buffer,1,20+font_len+1+text_len+1,fd);
  fflush(fd);
}

