/* X/Motif CMN previewer

LINUX: gcc xcmnw.c -o xcmnw -I/usr/X11R6/include -L/usr/X11R6/lib -lXm -lXp -lXt -lXext -lX11 -lm
SGI: cc xcmnw.c -o xcmnw -lXm -lXt -lX11
SUN: gcc xcmnw.c -o xcmnw -lXm -lXt -lX11
OSX: gcc xcmnw.c -o xcmnw -I/usr/X11R6/include /usr/local/lib/libXm.a /usr/X11R6/lib/libXt.a /usr/X11R6/lib/libXp.a -L/usr/X11R6/lib -lSM -lICE -lXext -lX11 -framework CoreFoundation

*/

#include <ctype.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <Xm/XmAll.h>

typedef struct {
  int op;
  int filled; 
  int size;
  int even_odd;
  int *idat;
  XPoint *pdat;
  char *font;
  char *str;
} pt;

static pt** pts = NULL;

static Display *dpy;
static Drawable win;
static GC gc;
static Colormap cmap;
static XColor tmp_color;

static int points = 0;
static int current_point = 0;

#define PTS_INCREMENT 1024

#define CLEAR 0
#define LINE 1
#define TEXT 2
#define RECTANGLE 3
#define POLYGON 4
#define CIRCLE 5
#define QUIT 6
#define COLOR 7
#define LINE_WIDTH 8
#define LINE_STYLE 9
#define FILL_STYLE 10

static XtAppContext app;     
static Widget shell,pane;
static Arg args[32];
static XGCValues gv;
static int n;
static int scr;
static Atom wm_delete_window;
static int grf_r = 0, grf_g = 0, grf_b = 0;
static Pixel grf_color, black, white;
static XFontStruct *grf_fontstruct;

static char *grf_font = NULL;
static int grf_font_size = 0;


static void draw_score (void);

static void resize(Widget w, XtPointer clientData, XtPointer callData)
{
  if (current_point) draw_score();
}

static void stop_running(Widget w, XtPointer clientData, XtPointer callData)
{
  current_point = 0;
}

static char *copy_string (char *str)
{
  char *newstr = NULL;
  if (str)
    {
      newstr = (char *)calloc(strlen(str) + 1, sizeof(char));
      strcpy(newstr, str);
    }
  return(newstr);
}

static pt *next_pt(void)
{
  int i;
  pt *npt;
  if (points <= current_point)
    {
      if (!pts) 
	pts = (pt **)calloc(PTS_INCREMENT, sizeof(pt *));
      else 
	{
	  pts = (pt **)realloc(pts, (points + PTS_INCREMENT) * sizeof(pt *));
	  for (i = points; i < (points + PTS_INCREMENT); i++) pts[i] = NULL;
	}
      points += PTS_INCREMENT;
    }
  if (!(pts[current_point])) 
    {
      pts[current_point] = (pt *)calloc(1, sizeof(pt));
      npt = pts[current_point];
    }
  else
    {
      npt = pts[current_point];
      if (npt->str) {free(npt->str); npt->str = NULL;}
      if (npt->idat) {free(npt->idat); npt->idat = NULL;}
      if (npt->font) {free(npt->font); npt->font = NULL;}
      if (npt->pdat) {free(npt->pdat); npt->pdat = NULL;}
    }
  current_point++;
  return(npt);
}

static void x_clear_window (void)
{
  XClearWindow(dpy, win);
  current_point = 0;
  if (grf_color != black)
    {
      grf_r = 0;
      grf_g = 0;
      grf_b = 0;
      grf_color = black;
      XSetForeground(dpy, gc, black);
    }
  grf_font = NULL;
  grf_font_size = 0;
}

static int fonts_loaded = 0;
static char *font_names[256];
static XFontStruct *font_structs[256];

static set_font (char *font,  int size)
{
  int i;
  if ((size != grf_font_size) || (strcmp(font, grf_font)))
    {
      /* look to see if we already have it loaded */
      grf_font_size = size;
      if (grf_font) free(grf_font);
      grf_font = copy_string(font);
      for (i = 0; i < fonts_loaded; i++)
	{
	  if ((strcmp(font, font_names[i])) == 0) 
	    {
	      grf_fontstruct = font_structs[i];
	      XSetFont(dpy, gc, grf_fontstruct->fid);
	      return;
	    }
	}
      grf_fontstruct = XLoadQueryFont(dpy, font);
      if (grf_fontstruct)
	{
	  font_names[fonts_loaded] = copy_string(font);
	  font_structs[fonts_loaded] = grf_fontstruct;
	  fonts_loaded++;
	}
      else grf_fontstruct = XLoadQueryFont(dpy, "fixed");
      if (grf_fontstruct)
	XSetFont(dpy, gc, grf_fontstruct->fid);
    }
}

static void x_color(int *arr)
{
  /* r g b */
  int r, g, b;
  pt *npt;
  r = arr[0];
  g = arr[1];
  b = arr[2];
  if ((grf_r != r) || (grf_g != g) || (grf_b != b))
    {
      if (r == 256) tmp_color.red = 65535; else tmp_color.red = r * 256;
      if (g == 256) tmp_color.green = 65535; else tmp_color.green = g * 256;
      if (b == 256) tmp_color.blue = 65535; else tmp_color.blue = b * 256;
      grf_r = r;
      grf_g = g;
      grf_b = b;
      if ((r) || (g) || (b))
	{
	  if ((r == 1) && (g == 1) && (b == 1))
	    {
	      grf_color = white;
	    }
	  else
	    {
	      XAllocColor(dpy, cmap, &tmp_color);
	      grf_color = tmp_color.pixel;
	    }
	}
      else
	{
	  grf_color = black;
	}
      XSetForeground(dpy, gc, grf_color);
      npt = next_pt();
      npt->op = COLOR;
      npt->idat = (int *)calloc(1, sizeof(int));
      npt->idat[0] = (int)grf_color;
    }
}

static int current_line_width, current_line_style, current_fill_style;

static void x_line_width(int *arr)
{
  pt *npt;
  current_line_width = arr[0];
  XSetLineAttributes(dpy, gc, current_line_width, current_line_style, CapButt, JoinMiter);
  npt = next_pt();
  npt->op = LINE_WIDTH;
  npt->idat = (int *)calloc(1, sizeof(int));
  npt->idat[0] = current_line_width;
}

static void x_line_style(int *arr)
{
  pt *npt;
  if (arr[0]) current_line_style = LineOnOffDash; else current_line_style = LineSolid;
  XSetLineAttributes(dpy, gc, current_line_width, current_line_style, CapButt, JoinMiter);
  npt = next_pt();
  npt->op = LINE_STYLE;
  npt->idat = (int *)calloc(1, sizeof(int));
  npt->idat[0] = current_line_style;
}

static void x_fill_style(int *arr)
{
  pt *npt;
  if (arr[0]) current_fill_style = FillStippled; else current_fill_style = FillSolid;
  XSetFillStyle(dpy, gc, current_fill_style);
  npt = next_pt();
  npt->op = FILL_STYLE;
  npt->idat = (int *)calloc(1, sizeof(int));
  npt->idat[0] = current_fill_style;
}

static void x_line (int *arr)
{
  /* x0 y0 x1 y1 */
  pt *npt;
  int i;
  XDrawLine(dpy, win, gc, arr[0], arr[1], arr[2], arr[3]);
  npt = next_pt();
  npt->op = LINE;
  npt->idat = (int *)calloc(4, sizeof(int));
  for (i = 0; i < 4; i++) npt->idat[i] = arr[i];
}

static void x_rectangle(int filled,  int *arr)
{
  /* x0 y0 width height */
  int i;
  pt *npt;
  if (filled)
    XFillRectangle(dpy, win, gc, arr[0], arr[1], arr[2], arr[3]);
  else XDrawRectangle(dpy, win, gc, arr[0], arr[1], arr[2], arr[3]);
  npt = next_pt(); 
  npt->op = RECTANGLE;
  npt->filled = filled;
  npt->idat = (int *)calloc(4, sizeof(int));
  for (i = 0; i < 4; i++) npt->idat[i] = arr[i];
}

static void x_polygon(int filled,  int even_odd,  int size,  int *arr)
{
  int i, j;
  pt *npt;
  npt = next_pt(); 
  npt->op = POLYGON;
  npt->filled = filled;
  npt->even_odd = even_odd;
  npt->size = size;
  npt->pdat = (XPoint *)calloc(size, sizeof(XPoint));
  for (i = 0, j = 0; i < size; i++, j += 2) 
    {
      npt->pdat[i].x = arr[j];
      npt->pdat[i].y = arr[j+1];
    }
  if ((filled) && (even_odd)) XSetFillRule(dpy, gc, EvenOddRule);
  if (filled)
    XFillPolygon(dpy, win, gc, npt->pdat, size, Complex, CoordModeOrigin);
  else XDrawLines(dpy, win, gc, npt->pdat, size, CoordModeOrigin);
  if ((filled) && (even_odd)) XSetFillRule(dpy, gc, WindingRule);
}

static void x_circle(int filled,  int *arr)
{
  pt *npt;
  int i, x0, y0, diameter, radius, a1, a2;
  /* arr = x, y:(center),  radius,  but in X its the upper left corner + width/height */
  /* also in X,  2nd ang is relative to first */
  radius = arr[2];
  diameter = radius * 2;
  x0 = arr[0] - radius;
  y0 = arr[1] - radius;
  a1 = arr[3];
  a2 = arr[4] - arr[3];
  if (filled)
    XFillArc(dpy, win, gc, x0, y0, diameter, diameter, a1 * 64, a2 * 64);
  else XDrawArc(dpy, win, gc, x0, y0, diameter, diameter, a1 * 64, a2 * 64);
  npt = next_pt(); 
  npt->op = CIRCLE;
  npt->filled = filled;
  npt->idat = (int *)calloc(5, sizeof(int));
  npt->idat[0] = x0;
  npt->idat[1] = y0;
  npt->idat[2] = diameter;
  npt->idat[3] = a1;
  npt->idat[4] = a2;
}

void x_text(int font_size,  int *arr,  char *font,  char *text)
{
  pt *npt;
  if (text)
    {
      set_font(font, font_size);
      XDrawString(dpy, win, gc, arr[0], arr[1], text, strlen(text));
      npt = next_pt(); 
      npt->op = TEXT;
      npt->size = font_size;
      npt->font = copy_string(font);
      npt->str = copy_string(text);
      npt->idat = (int *)calloc(2, sizeof(int));
      npt->idat[0] = arr[0];
      npt->idat[1] = arr[1];
    }
}

static void draw_score (void)
{
  int i;
  pt *npt;
  XClearWindow(dpy, win);
  for (i = 0; i < current_point; i++)
    {
      npt = pts[i];
      switch (npt->op)
	{
	case LINE: 
	  XDrawLine(dpy, win, gc, npt->idat[0], npt->idat[1], npt->idat[2], npt->idat[3]); 
	  break;
	case RECTANGLE: 
	  if (npt->filled)
	    XFillRectangle(dpy, win, gc, npt->idat[0], npt->idat[1], npt->idat[2], npt->idat[3]);
	  else XDrawRectangle(dpy, win, gc, npt->idat[0], npt->idat[1], npt->idat[2], npt->idat[3]);
	  break;
	case CIRCLE:
	  if (npt->filled)
	    XFillArc(dpy, win, gc, npt->idat[0], npt->idat[1], npt->idat[2], npt->idat[2], npt->idat[3] * 64, npt->idat[4] * 64);
	  else XDrawArc(dpy, win, gc, npt->idat[0], npt->idat[1], npt->idat[2], npt->idat[2], npt->idat[3] * 64, npt->idat[4] * 64);
	  break;
	case TEXT: 
	  set_font(npt->font, npt->size);
	  XDrawString(dpy, win, gc, npt->idat[0], npt->idat[1], npt->str, strlen(npt->str));
	  break;
	case POLYGON:
	  if ((npt->filled) && (npt->even_odd)) XSetFillRule(dpy, gc, EvenOddRule);
	  if (npt->filled)
	    XFillPolygon(dpy, win, gc, npt->pdat, npt->size, Complex, CoordModeOrigin);
	  else XDrawLines(dpy, win, gc, npt->pdat, npt->size, CoordModeOrigin);
	  if ((npt->filled) && (npt->even_odd)) XSetFillRule(dpy, gc, WindingRule);
	  break;
	case COLOR:
	  XSetForeground(dpy, gc, (Pixel)(npt->idat[0]));
	  break;
	case LINE_WIDTH:
	  current_line_width = npt->idat[0];
	  XSetLineAttributes(dpy, gc, current_line_width, current_line_style, CapButt, JoinMiter);
	  break;
	case LINE_STYLE:
	  current_line_style = npt->idat[0];
	  XSetLineAttributes(dpy, gc, current_line_width, current_line_style, CapButt, JoinMiter);
	  break;
	case FILL_STYLE:
	  current_fill_style = npt->idat[0];
	  XSetFillStyle(dpy, gc, current_fill_style);
	  break;
	}
    }
}

static char *buffer = NULL;
static int buffer_size = 0;

static void flush_it(int *fd)
{
  int n;
  n = buffer_size;
  while (n == buffer_size) {n = read(*fd, buffer, buffer_size);}
}

static void GetCmnString (XtPointer clientData,  int *fd,  XtInputId *id)
{
  int bytes, size, n, bytes_read;
  int *ibuf;
  char *font, *text, *tmpbuf;
  n = read(*fd, buffer, 4);
  if (n <= 0) exit(0); /* probably a dangling connection */
  if (n != 4) {fprintf(stderr, "got %d bytes reading len?\n", n); return;}
  size = (*((int *)buffer));
  if (size == 0) {fprintf(stderr, "must have dropped a command\n"); flush_it(fd); return;}
  if (size > buffer_size)
    {
      if (size > 65536) {fprintf(stderr, "attempt to set buffer size to %d = %s!\n", size, buffer); return;}
      buffer_size = size * 2;
      buffer = (char *)realloc(buffer, buffer_size * sizeof(char));
    }
  bytes_read = 0;
  tmpbuf = buffer;
  while (bytes_read < size)
    {
      n = read(*fd, tmpbuf, size);
      bytes_read += n;
      tmpbuf = (char *)(buffer + bytes_read);
    }
  switch (*((int *)buffer))
    {
    case CLEAR:
      x_clear_window();
      break;
    case LINE:
      ibuf = (int *)(buffer + 4);
      x_line(ibuf);
      break;
    case RECTANGLE: 
      ibuf = (int *)(buffer + 8);
      x_rectangle((*((int *)(buffer + 4))), ibuf);
      break;
    case POLYGON:
      ibuf = (int *)(buffer + 16);
      x_polygon((*((int *)(buffer + 4))), 
		(*((int *)(buffer + 8))), 
		(*((int *)(buffer + 12))), 
		ibuf);
      break;
    case CIRCLE:
      ibuf = (int *)(buffer + 8);
      x_circle((*((int *)(buffer + 4))), ibuf);
      break;
    case TEXT:
      size = (*((int *)(buffer + 4)));
      ibuf = (int *)(buffer + 8);
      font = copy_string((char *)(buffer + 16));
      text = copy_string((char *)(buffer + 16 + strlen(font) + 1));
      x_text(size, ibuf, font, text);
      break;
    case COLOR:
      ibuf = (int *)(buffer + 4);
      x_color(ibuf);
      break;
    case LINE_WIDTH:
      ibuf = (int *)(buffer + 4);
      x_line_width(ibuf);
      break;
    case LINE_STYLE:
      ibuf = (int *)(buffer + 4);
      x_line_style(ibuf);
      break;
    case FILL_STYLE:
      ibuf = (int *)(buffer + 4);
      x_fill_style(ibuf);
      break;
    case QUIT:
      exit(0);
      break;
    default: 
      flush_it(fd); 
      break;
    }
}

int main(int argc,  char **argv)
{
  buffer_size = 8192;
  buffer = (char *)calloc(buffer_size, sizeof(char));

  shell = XtVaOpenApplication(&app, "mainshell", NULL, 0, &argc, argv, NULL, applicationShellWidgetClass, 
			      XmNallowShellResize, TRUE, 
			      XmNwidth, 500, XmNheight, 500, 
			      NULL);

  dpy=XtDisplay(shell);
  scr=DefaultScreen(dpy);
  cmap=DefaultColormap(dpy, scr);

  n=0;
  XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetArg(args[n], XmNallowResize, TRUE); n++;
  black = BlackPixel(dpy, scr);
  white = WhitePixel(dpy, scr);
  XtSetArg(args[n], XmNbackground, white); n++;
  XtSetArg(args[n], XmNforeground, black); n++;
  pane = XtCreateManagedWidget("mainpane", xmDrawingAreaWidgetClass, shell, args, n);
  
  XtRealizeWidget(shell);
  
  gv.function = GXcopy;
  XtVaGetValues(pane, XmNbackground,  &gv.background, XmNforeground,  &gv.foreground, NULL);
  gc = XtGetGC(pane,  GCBackground | GCForeground | GCFunction,  &gv);
  XSetFillRule(dpy, gc, WindingRule);
  grf_r = 0;
  grf_g = 0;
  grf_b = 0;
  current_point = 0;
  grf_color = black;

  win = XtWindow(pane);
  
  XtAddCallback(pane, XmNresizeCallback, resize, NULL);
  XtAddCallback(pane, XmNexposeCallback, resize, NULL);
  
  wm_delete_window = XmInternAtom(dpy, "WM_DELETE_WINDOW", FALSE);
  XmAddWMProtocolCallback(shell, wm_delete_window, stop_running, NULL);
  XtAppAddInput(app, fileno(stdin), (XtPointer)XtInputReadMask, GetCmnString, NULL);
  XtAppMainLoop(app);
}

/* us->cmn only if we stop via local menu */

